/*
    Project: TCTEMP
    Version: V1.7
    File   : tc_crypt.c

    Copyright (c) 2006-2008 Author of TCTEMP. All rights reserved.
    Public Key Fingerprint: 75EB 6BC2 01B7 F6E7 4BD7 CC58 4A5F C393 19EE 6E69

    Covered by TrueCrypt License 2.3 the full text of which is contained
    in the file License.txt included in TCTEMP binary and source code
    distribution archives.

    The copyright notice of this source code file and the public key
    fingerprint must be retained in full.
*/

#include "tc_priv.h"

/*----------------------------------------------------------------------------*/
/* Functions                                                                  */

#define SETUP(idx, cipher, key, key_length) \
    if (cipher##_desc.setup(key, key_length, \
      0, &pCtx->sk[idx]) != CRYPT_OK) goto error; \
      pCtx->pdCipher[idx] = &cipher##_desc;

BOOL __stdcall TC_KeyIsWeak(PTC_KEY pKey, unsigned uCipher)
{
    const unsigned __int64 *pullK;

    pullK = (const unsigned __int64 *) pKey->rgbSecondaryKey;

    if (pullK[0] == 0 && pullK[1] == 0) {
        /* weak secondary key */
        return TRUE;
    }

#ifdef DES
    if (uCipher == TC_TRIPLEDES) {
        return TC_TripleDESKeyIsWeak(pKey->rgbMasterKey);
    }
#endif

    return FALSE;
}

BOOL __stdcall TCE_Init(PTCE_CONTEXT pCtx, PTC_KEY pKey, unsigned uCipher)
{
    memset(pCtx, '\0', sizeof(*pCtx));
    pCtx->rgbI[7] = 1;

    switch(uCipher) {
#ifdef RIJNDAEL
    case TC_AES:
        SETUP(0, aes, pKey->rgbMasterKey, 32);
        break;
#endif
#ifdef BLOWFISH
    case TC_BLOWFISH:
        SETUP(0, tcblowfish, pKey->rgbMasterKey, 56);
        break;
#endif
#ifdef CAST5
    case TC_CAST5:
        SETUP(0, cast5, pKey->rgbMasterKey, 16);
        break;
#endif
#ifdef SERPENT
    case TC_SERPENT:
        SETUP(0, serpent, pKey->rgbMasterKey, 32);
        break;
#endif
#ifdef DES
    case TC_TRIPLEDES:
        SETUP(0, des3, pKey->rgbMasterKey, 24);
        break;
#endif
#ifdef TWOFISH
    case TC_TWOFISH:
        SETUP(0, twofish, pKey->rgbMasterKey, 32);
        break;
#endif
#if defined(RIJNDAEL) && defined(TWOFISH)
    case TC_AES_TWOFISH:
        SETUP(0, twofish, pKey->rgbMasterKey, 32);
        SETUP(1, aes, pKey->rgbMasterKey + 32, 32);
        break;
#endif
#if defined(RIJNDAEL) && defined(TWOFISH) && defined(SERPENT)
    case TC_AES_TWOFISH_SERPENT:
        SETUP(0, serpent, pKey->rgbMasterKey, 32);
        SETUP(1, twofish, pKey->rgbMasterKey + 32, 32);
        SETUP(2, aes, pKey->rgbMasterKey + 64, 32);
        break;
#endif
#if defined(SERPENT) && defined(RIJNDAEL)
    case TC_SERPENT_AES:
        SETUP(0, aes, pKey->rgbMasterKey, 32);
        SETUP(1, serpent, pKey->rgbMasterKey + 32, 32);
        break;
#endif
#if defined(SERPENT) && defined(TWOFISH) && defined(RIJNDAEL)
    case TC_SERPENT_TWOFISH_AES:
        SETUP(0, aes, pKey->rgbMasterKey, 32);
        SETUP(1, twofish, pKey->rgbMasterKey + 32, 32);
        SETUP(2, serpent, pKey->rgbMasterKey + 64, 32);
        break;
#endif
#if defined(TWOFISH) && defined(SERPENT)
    case TC_TWOFISH_SERPENT:
        SETUP(0, serpent, pKey->rgbMasterKey, 32);
        SETUP(1, twofish, pKey->rgbMasterKey + 32, 32);
        break;
#endif
    default:
        return FALSE;
    }

    memcpy(pCtx->rgbSecondaryKey, pKey->rgbSecondaryKey,
      sizeof(pCtx->rgbSecondaryKey));

    return TRUE;

error:
    TCE_Close(pCtx);
    return FALSE;
}

BOOL __stdcall TCE_SetIndex(PTCE_CONTEXT pCtx,
  unsigned __int64 ullIndex)
{
    const unsigned __int8 *pbSrc  = (const unsigned __int8 *)&ullIndex;
    unsigned __int8       *pbDest = pCtx->rgbI;

    if (ullIndex == 0) {
        return FALSE;
    }

    pbDest[0] = pbSrc[7];
    pbDest[1] = pbSrc[6];
    pbDest[2] = pbSrc[5];
    pbDest[3] = pbSrc[4];
    pbDest[4] = pbSrc[3];
    pbDest[5] = pbSrc[2];
    pbDest[6] = pbSrc[1];
    pbDest[7] = pbSrc[0];

    return TRUE;
}

BOOL __stdcall TCE_SetSectorNumber(PTCE_CONTEXT pCtx,
  unsigned __int64 ullSectorNumber)
{
    if (pCtx->pdCipher[0]->block_length == 8) {
        ullSectorNumber *= 64;
        return TCE_SetIndex(pCtx, ullSectorNumber + 1);
    } else {
        ullSectorNumber *= 32;
        return TCE_SetIndex(pCtx, ullSectorNumber + 1);
    }
}

BOOL __stdcall TCE_Encrypt(PTCE_CONTEXT pCtx, void *pDest, size_t cbDestSize)
{
    unsigned __int8       *pbDest;
    unsigned __int8       *pbDestEnd;
    unsigned              i;

    TC_ASSERT(pCtx);
    TC_ASSERT(pCtx->pdCipher[0]);
    TC_ASSERT(pDest);
    TC_ASSERT(cbDestSize != 0);
    TC_ASSERT(cbDestSize % 16 == 0);

    pbDest = (unsigned __int8 *)pDest;
    pbDestEnd = pbDest + cbDestSize;

    if (pCtx->pdCipher[0]->block_length == 8) {

        TC_ASSERT(pCtx->pdCipher[1] == NULL);
        TC_ASSERT(pCtx->pdCipher[2] == NULL);

        while (pbDest != pbDestEnd) {

            GfMul64(pCtx->rgbT, pCtx->rgbSecondaryKey, pCtx->rgbI);
            Xor64(pbDest, pCtx->rgbT);
            if ((pCtx->pdCipher[0]->ecb_encrypt(pbDest, pbDest,
              &pCtx->sk[0])) != CRYPT_OK) {
                return FALSE;
            }
            Xor64(pbDest, pCtx->rgbT);

            for (i = 8; i--; ) {
                if (++pCtx->rgbI[i]) break;
            }

            pbDest += 8;
        }

    } else {

        TC_ASSERT(pCtx->pdCipher[0]->block_length == 16);

        while (pbDest != pbDestEnd) {

            GfMul128(pCtx->rgbT, pCtx->rgbSecondaryKey, pCtx->rgbI);
            Xor128(pbDest, pCtx->rgbT);
            if ((pCtx->pdCipher[0]->ecb_encrypt(pbDest, pbDest,
              &pCtx->sk[0])) != CRYPT_OK) {
                return FALSE;
            }
            if (pCtx->pdCipher[1]) {
                if ((pCtx->pdCipher[1]->ecb_encrypt(pbDest, pbDest,
                  &pCtx->sk[1])) != CRYPT_OK) {
                    return FALSE;
                }
                if (pCtx->pdCipher[2]) {
                    if ((pCtx->pdCipher[2]->ecb_encrypt(pbDest, pbDest,
                    &pCtx->sk[2])) != CRYPT_OK) {
                        return FALSE;
                    }
                }
            }
            Xor128(pbDest, pCtx->rgbT);

            for (i = 8; i--; ) {
                if (++pCtx->rgbI[i]) break;
            }

            pbDest += 16;
        }
    }

    return TRUE;
}

BOOL __stdcall TCE_Decrypt(PTCE_CONTEXT pCtx, void *pDest, size_t cbDestSize)
{
    unsigned __int8       *pbDest;
    unsigned __int8       *pbDestEnd;
    unsigned              i;

    TC_ASSERT(pCtx);
    TC_ASSERT(pCtx->pdCipher[0]);
    TC_ASSERT(pDest);
    TC_ASSERT(cbDestSize != 0);
    TC_ASSERT(cbDestSize % 16 == 0);

    pbDest = (unsigned __int8 *)pDest;
    pbDestEnd = pbDest + cbDestSize;

    if (pCtx->pdCipher[0]->block_length == 8) {

        TC_ASSERT(pCtx->pdCipher[1] == NULL);
        TC_ASSERT(pCtx->pdCipher[2] == NULL);

        while (pbDest != pbDestEnd) {

            GfMul64(pCtx->rgbT, pCtx->rgbSecondaryKey, pCtx->rgbI);
            Xor64(pbDest, pCtx->rgbT);
            if ((pCtx->pdCipher[0]->ecb_decrypt(pbDest, pbDest,
              &pCtx->sk[0])) != CRYPT_OK) {
                return FALSE;
            }
            Xor64(pbDest, pCtx->rgbT);

            for (i = 8; i--; ) {
                if (++pCtx->rgbI[i]) break;
            }
            pbDest += 8;
        }

    } else {

        TC_ASSERT(pCtx->pdCipher[0]->block_length == 16);

        while (pbDest != pbDestEnd) {

            GfMul128(pCtx->rgbT, pCtx->rgbSecondaryKey, pCtx->rgbI);
            Xor128(pbDest, pCtx->rgbT);
            if (pCtx->pdCipher[2]) {
                if ((pCtx->pdCipher[2]->ecb_decrypt(pbDest, pbDest,
                &pCtx->sk[2])) != CRYPT_OK) {
                    return FALSE;
                }
            }
            if (pCtx->pdCipher[1]) {
                if ((pCtx->pdCipher[1]->ecb_decrypt(pbDest, pbDest,
                  &pCtx->sk[1])) != CRYPT_OK) {
                    return FALSE;
                }
            }
            if ((pCtx->pdCipher[0]->ecb_decrypt(pbDest, pbDest,
              &pCtx->sk[0])) != CRYPT_OK) {
                return FALSE;
            }
            Xor128(pbDest, pCtx->rgbT);

            for (i = 8; i--; ) {
                if (++pCtx->rgbI[i]) break;
            }

            pbDest += 16;
        }
    }

    return TRUE;
}

void __stdcall TCE_Close(PTCE_CONTEXT pCtx)
{
    if (pCtx->pdCipher[0]) {
        pCtx->pdCipher[0]->done(&pCtx->sk[0]);
    }
    if (pCtx->pdCipher[1]) {
        pCtx->pdCipher[1]->done(&pCtx->sk[1]);
    }
    if (pCtx->pdCipher[2]) {
        pCtx->pdCipher[2]->done(&pCtx->sk[2]);
    }
    memset(pCtx, '\0', sizeof(*pCtx));
}
